#include "mpi.h"

#ifndef ___dcomplex_definition___
typedef struct { double r,i; } dcomplex;
#define ___dcomplex_definition___ 
#endif

#include <fftw3.h> 


/*******************************************************
  openfft_initialize is used to prepare data for the transformation 
     Input: 3 dimensions of data: N1, N2, N3, measure_time and print_memory (0: disabled, 1: enabled)
     Output: arrays allocated and variables initialized
             My_Max_NumGrid: the maximum number of grid points allocated to a process
             My_NumGrid_In : the number of grid points allocated to a process for the first transformation
             My_Index_In   : the indexes of grid points allocated to a process for the first transformation
             My_NumGrid_Out: the number of grid points allocated to a process for the last transformation
             My_Index_Out  : the indexes of grid points allocated to a process for the last transformation
******************************************************/
double openfft_initialize(int N1, int N2, int N3, int measure_time, int print_memory,
			  int *My_Max_NumGrid, 
			  int *My_NumGrid_In, int *My_Index_In,
			  int *My_NumGrid_Out, int *My_Index_Out);


/*******************************************************
  openfft_execute is used to transform the data 
     Input: Rhor
     Output: Rhok
******************************************************/
double openfft_execute(dcomplex *Rhor, dcomplex *Rhok);  

/*******************************************************
  openfft_finalize is used to clean up the calculation 
******************************************************/
double openfft_finalize();  

/*******************************************************
  OpenFFT_dtime is used to measure time 
******************************************************/
void openfft_dtime(double *time);
 

/*******************************************************
 int *Num_Snd_Grid_AB2CA

  Num_Snd_Grid_AB2CA gives the number of grid points
  sent from the AB partition to the CA partition.
  size: Num_Snd_Grid_AB2CA[numprocs]
*******************************************************/
int *Num_Snd_Grid_AB2CA;

/*******************************************************
 int *Num_Rcv_Grid_AB2CA

  Num_Rcv_Grid_AB2CA gives the number of grid points
  received in the CA partition and sent from the AB 
  partition.
  size: Num_Rcv_Grid_AB2CA[numprocs]
*******************************************************/
int *Num_Rcv_Grid_AB2CA;

/*******************************************************
 int *Num_Snd_Grid_CA2CB

  Num_Snd_Grid_CA2CB gives the number of grid points
  sent from the CA partition to the CB partition.
  size: Num_Snd_Grid_CA2CB[numprocs]
*******************************************************/
int *Num_Snd_Grid_CA2CB;

/*******************************************************
 int *Num_Rcv_Grid_CA2CB

  Num_Rcv_Grid_CA2CB gives the number of grid points
  received in the CB partition and sent from the CA
  partition.
  size: Num_Rcv_Grid_CA2CB[numprocs]
*******************************************************/
int *Num_Rcv_Grid_CA2CB;

/*******************************************************
 int **Index_Snd_Grid_AB2CA

  Index_Snd_Grid_AB2CA gives index, BN_AB in the partition 
  AB associated with the grid points sent to ID.
  size: Index_Snd_Grid_AB2CA[numprocs][Num_Snd_Grid_AB2CA[ID]]
*******************************************************/
int **Index_Snd_Grid_AB2CA;

/*******************************************************
 int **Index_Rcv_Grid_AB2CA

  Index_Rcv_Grid_AB2CA gives index, BN_AB in the partition 
  AB associated with the grid point received from ID.
  size: Index_Rcv_Grid_AB2CA[numprocs][Num_Rcv_Grid_AB2CA[ID]]
*******************************************************/
int **Index_Rcv_Grid_AB2CA;

/*******************************************************
 int **Index_Snd_Grid_CA2CB

  Index_Snd_Grid_CA2CB gives index, BN_CA in the partition 
  CA associated with the grid points sent to ID.
  size: Index_Snd_Grid_CA2CB[numprocs][Num_Snd_Grid_CA2CB[ID]]
*******************************************************/
int **Index_Snd_Grid_CA2CB;

/*******************************************************
 int **Index_Rcv_Grid_CA2CB

  Index_Rcv_Grid_CA2CB gives index, BN_CA in the partition 
  CA associated with the grid points received from ID.
  size: Index_Rcv_Grid_CA2CB[numprocs][Num_Rcv_Grid_CA2CB[ID]]
*******************************************************/
int **Index_Rcv_Grid_CA2CB;

/*******************************************************
  int *ID_NN_AB2CA_S;
  int *ID_NN_AB2CA_R;

  global process ID used for sending and receiving data
  in MPI commucation (AB to CA).
*******************************************************/
int *ID_NN_AB2CA_S;
int *ID_NN_AB2CA_R;

/*******************************************************
  int *GP_AB2CA_S;
  int *GP_AB2CA_R;

  starting index to data used for sending and receiving data
  in MPI commucation (AB to CA).

*******************************************************/
int *GP_AB2CA_S;
int *GP_AB2CA_R;

/*******************************************************
  int *ID_NN_CA2CB_S;
  int *ID_NN_CA2CB_R;

  global process ID used for sending and receiving data
  in MPI commucation (CA to CB).

*******************************************************/
int *ID_NN_CA2CB_S;
int *ID_NN_CA2CB_R;

/*******************************************************
  int *GP_CA2CB_S;
  int *GP_CA2CB_R;

  starting index to data used for sending and receiving data
  in MPI commucation (CA to CB).

*******************************************************/
int *GP_CA2CB_S;
int *GP_CA2CB_R;

double OFFT_timers[4];

fftw_complex *OFFT_in, *OFFT_out;
fftw_plan OFFT_p_A,OFFT_p_B,OFFT_p_C;

int OFFT_Ngrid1,OFFT_Ngrid2,OFFT_Ngrid3;
int OFFT_My_Max_NumGrid,Max_OneD_Grids;
int My_NumGrid_AB,My_NumGrid_CB,My_NumGrid_CA;
int OFFT_My_NumGrid_In,OFFT_My_NumGrid_Out;
int Max_Num_Snd_Grid_AB2CA;
int Max_Num_Rcv_Grid_AB2CA;
int Max_Num_Snd_Grid_CA2CB;
int Max_Num_Rcv_Grid_CA2CB;
int NN_AB2CA_S,NN_AB2CA_R,NN_CA2CB_S,NN_CA2CB_R;
int OFFT_alloc_first,OFFT_alloc_first1,OFFT_measure_time,OFFT_print_memory;
int size_Index_Snd_Grid_AB2CA,size_Index_Rcv_Grid_AB2CA,size_Index_Snd_Grid_CA2CB,size_Index_Rcv_Grid_CA2CB;

