!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!  sample05.f90
!   This program is used for illustrating how to call OpenFFT from a Fortran user program.
!   It can be executed with an arbitrary number of processes. 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

program sample05

  use MPI
  use, intrinsic :: iso_c_binding
  implicit none
  include 'openfft.fi'  
  
  integer(C_INT), parameter :: N1=64, N2=64, N3=64
  
  integer(C_INT), parameter :: NTEST = 10 
  
  complex(C_DOUBLE_COMPLEX), allocatable, dimension(:) :: Rhor, Rhok, in1

  integer(C_INT), dimension(6) :: My_Index_In,My_Index_Out
  
  real(C_DOUBLE) :: nn,flops
  integer(C_INT) :: ierror, nproc, nrank, i,j,k,m,mpi_world,measure_time,print_memory,&
       My_Max_NumGrid,My_NumGrid_In,My_NumGrid_Out 
  real(C_DOUBLE) :: t1, t2, t3 ,t4
  real(C_DOUBLE) :: err, err_all
  
  call MPI_INIT(ierror)
  call MPI_COMM_SIZE(MPI_COMM_WORLD, nproc, ierror)
  call MPI_COMM_RANK(MPI_COMM_WORLD, nrank, ierror)

! Set whether to use the timing and print memory functions of OpenFFT or not. Default=0 (not use) 

  measure_time = 0
  print_memory = 0
  
! Initialize OpenFFT  

  t3 = openfft_initialize(%VAL(N1),%VAL(N2),%VAL(N3),%VAL(measure_time),%VAL(print_memory),&
       My_Max_NumGrid,My_NumGrid_In,My_Index_In,My_NumGrid_Out,My_Index_Out)

! Allocate local input and output arrays 

  allocate (Rhor(My_Max_NumGrid))
  allocate (Rhok(My_Max_NumGrid))
  allocate (in1(My_Max_NumGrid))

! Set local input 

  do i=1,My_NumGrid_In
     Rhor(i) = (0.1, 0.2)
  end do
  
  in1 = Rhor
  
  t2 = 0.0D0
  t4 = 0.0D0
  t3 = 0.0D0

! Repeat FFT transform for NTEST times 

  do m=1,NTEST
     t1 = MPI_WTIME()
     t3=openfft_execute(Rhor, Rhok)
     t2 = t2 + MPI_WTIME() - t1

! Re-set local input 

     Rhor = in1
  end do
  
! Average time of all processes */

  call MPI_ALLREDUCE(t2,t1,1,MPI_DOUBLE_PRECISION,MPI_SUM, &
       MPI_COMM_WORLD,ierror)

  t1 = t1 / real(nproc)
  
! for error checking */

  err = 0.0
  do i=1,My_NumGrid_Out
     err = err + real(Rhok(i)) + aimag(Rhok(i))
  end do

  call MPI_ALLREDUCE(err,err_all,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,ierror)
  err_all = err_all / real((N1*N2*N3))

! Calculate flops */

  nn = real((N1*N2*N3))
  nn = nn ** (1.0/3.0)
  flops = 5.0 * nn * log(nn) / log(2.0)
  flops = flops * 3.0 * nn**2  
  flops = flops / (t1/real(NTEST))

! Finalize OpenFFT 

  t3 = openfft_finalize()

! Free arrays 

  deallocate(Rhor,Rhok,in1)

! Print results */

  if (nrank==0) then
     write(*,*) '=======OpenFFT complex2complex========'
     write(*,*) 'N1=',N1,' N2=',N2,' N3=',N3,' numprocs=',nproc
     write(*,*) 'Time openfft_execute Total =  ', t1
     write(*,*) 'Gflops                     =  ', flops / real(1000**3)
     write(*,*) 'Err_all                    =  ', err_all 
  end if
  
  call MPI_FINALIZE(ierror)

end program sample05

